import { BaseEvent } from '#application/base/baseEvent'
import { UUID } from '#application/types'
import SpriteRepository from '#repositories/spriteRepository'
import { SpriteAction } from '#entities/spriteAction'
import sharp from 'sharp'
import fs from 'fs'

type Payload = {
  id: UUID
  name: string
  spriteActions: Array<{
    action: string
    sprites: string[]
    originX: number
    originY: number
    frameRate: number
  }>
}

export default class SpriteUpdateEvent extends BaseEvent {
  public listen(): void {
    this.socket.on('gm:sprite:update', this.handleEvent.bind(this))
  }

  private async handleEvent(data: Payload, callback: (success: boolean) => void): Promise<void> {
    try {
      if (!(await this.isCharacterGM())) return

      const spriteRepository = new SpriteRepository()
      const sprite = await spriteRepository.getById(data.id)
      if (!sprite) return callback(false)

      await spriteRepository.getEntityManager().populate(sprite, ['spriteActions'])

      // Update sprite in database
      await sprite.setName(data.name).save()

      const existingActions = sprite.getSpriteActions()

      // Remove all existing actions - we'll recreate the ones we need
      for (const existingAction of existingActions) {
        await spriteRepository.getEntityManager().removeAndFlush(existingAction)
      }

      // Create new actions
      for (const actionData of data.spriteActions) {
        const spriteAction = new SpriteAction()
        spriteAction.setSprite(sprite)
        sprite.getSpriteActions().add(spriteAction)

        spriteAction
          .setAction(actionData.action)
          .setSprites(actionData.sprites)
          .setOriginX(actionData.originX)
          .setOriginY(actionData.originY)
          .setFrameWidth(await this.calculateMaxWidth(actionData.sprites))
          .setFrameHeight(await this.calculateMaxHeight(actionData.sprites))
          .setFrameRate(actionData.frameRate)

        await spriteRepository.getEntityManager().persistAndFlush(spriteAction)

        // Generate sprite sheet for this action
        await this.generateSpriteSheet(actionData.sprites, sprite.getId(), actionData.action)
      }

      return callback(true)
    } catch (error) {
      console.error(`Error updating sprite ${data.id}:`, error)
      return callback(false)
    }
  }

  private async calculateMaxWidth(sprites: string[]): Promise<number> {
    if (!sprites.length) return 0

    const widths = await Promise.all(
      sprites.map(async (base64) => {
        const uri = base64.split(';base64,').pop()
        if (!uri) return 0

        const imgBuffer = Buffer.from(uri, 'base64')
        const image = await sharp(imgBuffer).metadata()
        return image.width ?? 0
      })
    )

    return Math.max(...widths)
  }

  private async calculateMaxHeight(sprites: string[]): Promise<number> {
    if (!sprites.length) return 0

    const heights = await Promise.all(
      sprites.map(async (base64) => {
        const uri = base64.split(';base64,').pop()
        if (!uri) return 0

        const imgBuffer = Buffer.from(uri, 'base64')
        const image = await sharp(imgBuffer).metadata()
        return image.height ?? 0
      })
    )

    return Math.max(...heights)
  }

  private async generateSpriteSheet(sprites: string[], spriteId: string, action: string): Promise<void> {
    try {
      // Skip if no sprites
      if (!sprites.length) return

      // Process all base64 images to buffers and get their dimensions
      const imageData = await Promise.all(
        sprites.map(async (base64) => {
          const uri = base64.split(';base64,').pop()
          if (!uri) throw new Error('Invalid base64 image')
          const buffer = Buffer.from(uri, 'base64')
          const metadata = await sharp(buffer).metadata()
          return { buffer, width: metadata.width ?? 0, height: metadata.height ?? 0 }
        })
      )

      // Find the largest dimensions
      const maxWidth = Math.max(...imageData.map((data) => data.width))
      const maxHeight = Math.max(...imageData.map((data) => data.height))

      // Extend all images to match the largest dimensions without resizing
      const resizedBuffers = await Promise.all(
        imageData.map(async ({ buffer, width, height }) => {
          // Calculate padding to center the sprite
          const topPadding = 0 // This is always 0. We don't need to calculate it
          const bottomPadding = Math.ceil((maxHeight - height) / 2)
          const leftPadding = Math.floor((maxWidth - width) / 2)
          const rightPadding = Math.ceil((maxWidth - width) / 2)

          return await sharp(buffer)
            .extend({
              top: 0,
              bottom: bottomPadding,
              left: leftPadding,
              right: rightPadding,
              background: { r: 0, g: 0, b: 0, alpha: 0 }
            })
            .toBuffer()
        })
      )

      // Create sprite sheet by combining images horizontally
      const spriteSheet = await sharp({
        create: {
          width: maxWidth * sprites.length,
          height: maxHeight,
          channels: 4,
          background: { r: 0, g: 0, b: 0, alpha: 0 }
        }
      })
        .composite(
          resizedBuffers.map((buffer, index) => ({
            input: buffer,
            left: index * maxWidth,
            top: 0
          }))
        )
        .png()
        .toBuffer()

      // Ensure directory exists
      const dir = `public/sprites/${spriteId}`
      await fs.promises.mkdir(dir, { recursive: true })

      // Save the sprite sheet
      await fs.promises.writeFile(`${dir}/${action}.png`, spriteSheet)
    } catch (error) {
      console.error('Error generating sprite sheet:', error)
      throw error
    }
  }
}
